# -*- coding: utf-8 -*-

"""

PySSE v0.2.6

Copyright (C) 2009 Emmanuel Benoît <ebenoit@ebenoit.info>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

import re
import sys
import traceback
from copy import deepcopy
from mod_python import apache
from lxml import etree




##############################################################################
# UTILITY CLASSES AND FUNCTIONS                                              #
##############################################################################

class SiteXML( object ):
    """
    This class serves as a placeholder for the XML tree of the site's main
    configuration. The file is only opened and parsed when the object's xml
    property is accessed.
    """

    # A few constants describing the various types of errors
    NoConfiguration     = 0
    ConfigurationError  = 1
    PageNotFound        = 2
    BadMethod           = 3
    PageError           = 4

    # HTTP status codes for the errors
    StatusCodes = {
            NoConfiguration     : apache.HTTP_OK ,
            ConfigurationError  : apache.HTTP_INTERNAL_SERVER_ERROR ,
            PageNotFound        : apache.HTTP_NOT_FOUND ,
            BadMethod           : apache.HTTP_METHOD_NOT_ALLOWED ,
            PageError           : apache.HTTP_INTERNAL_SERVER_ERROR ,
        }

    # Base names of XML files that contain the error pages
    ErrorFiles = {
            NoConfiguration     : 'no-config' ,
            ConfigurationError  : 'config-error' ,
            PageNotFound        : 'not-found' ,
            BadMethod           : 'bad-method' ,
            PageError           : 'page-error' ,
        }

    def __init__( self , data_path ):
        """
        Stores the file's path into the instance and mark the tree as absent
        from memory.
        """
        self.path = data_path
        self.tree = None

    @property
    def xml( self ):
        """
        This property corresponds to the XML tree for the site's main
        configuration file. If the file hadn't been opened yet, it is parsed.
        """
        if self.tree is not None:
            return self.tree
        self.tree = etree.parse( self.path + 'site.xml' )
        return self.tree

    def get( self , xpath_expr , xml_data = None , **kwargs ):
        """
        This method performs XPath queries using the correct namespaces. By
        default it queries the site.xml data, but the xml_data parameter
        allows it to query any element tree.
        """
        kwargs[ 'namespaces' ] = {
                'xhtml': 'http://www.w3.org/1999/xhtml' ,
                'ws': 'http://www.ebenoit.info/data'
            }
        if xml_data is None:
            target = self.xml
        else:
            target = xml_data
        return target.xpath( xpath_expr , **kwargs )

    def init_groups( self ):
        """This method initialises page groups."""
        groups = { }
        r = self.get( '/ws:data/ws:group' )
        for group_def in r:
            group = PageGroup( self , group_def )
            if group.id in groups:
                raise ValueError(
                    'duplicate group identifier "%s"' % group.id )
            groups[ group.id ] = group
        self.groups = groups
        self.page_groups = { }

    def get_groups( self , page ):
        """Identifies the list of groups a page belongs to."""
        if page not in self.page_groups:
            self.page_groups[ page ] = [ group.id
                    for group in self.groups.values( )
                        if group.in_group( page ) ]
        return self.page_groups[ page ]

    def _handle_error( self , error_type , error_data , lang ):
        """This private method generates the contents of an error page."""
        # Open the error data file, using the language's name if it is
        # provided and if the translation exists.
        if lang is None:
            data = None
        else:
            try:
                data = etree.parse( self.path + 'errors/'
                        + SiteXML.ErrorFiles[ error_type ]
                        + '-' + lang + '.xml' )
            except IOError:
                data = None
        if data is None:
            data = etree.parse( self.path + 'errors/'
                    + SiteXML.ErrorFiles[ error_type ] + '.xml' )
        # Load the error XSL transform
        error_transform = etree.XSLT( etree.parse(
                self.path + 'errors/error.xsl' ) )
        # Utility function that generates tags
        xhtml_ns = 'http://www.w3.org/1999/xhtml'
        ws_ns = 'http://www.ebenoit.info/data'
        nsmap = { None : xhtml_ns , 'ws' : ws_ns }
        def _make_tag( tag ):
            return etree.Element( '{%s}%s' % ( ws_ns , tag ) ,
                    nsmap = nsmap )
        # Generate the basic XML data for the error
        err_xml = _make_tag( 'error' )
        err_xhtml = _make_tag( 'error-xhtml' )
        err_xhtml.append( deepcopy( data.getroot() ) )
        err_xml.append( err_xhtml )
        # Generate XML data from the exception's data, if it exists
        if error_data is not None:
            e_type , e_val , e_tb = error_data
            exc_xml = _make_tag( 'exception' )
            t = _make_tag( 'type' )
            t.text = e_type.__name__
            exc_xml.append( t )
            t = _make_tag( 'text' )
            t.text = str( e_val )
            exc_xml.append( t )
            tb_xml = _make_tag( 'trace' )
            for tb_entry in traceback.extract_tb( e_tb ):
                tbe_xml = _make_tag( 'entry' )
                t = _make_tag( 'file' )
                t.text = tb_entry[ 0 ]
                tbe_xml.append( t )
                t = _make_tag( 'line' )
                t.text = str( tb_entry[ 1 ] )
                tbe_xml.append( t )
                t = _make_tag( 'function' )
                t.text = tb_entry[ 2 ]
                tbe_xml.append( t )
                tb_xml.append( tbe_xml )
            exc_xml.append( tb_xml )
            err_xml.append( exc_xml )
        return str( error_transform( err_xml ) )

    def error( self , request , error_type , error_data = None , lang = None ):
        """
        This method displays an error page then exits the request handler. If an
        error occurs while handling the error, a basic text error is displayed.
        """
        try:
            r_string = self._handle_error( error_type , error_data , lang )
            request.status = SiteXML.StatusCodes[ error_type ]
            set_content_type( request )
            request.write( r_string )
        except:
            try:
                e_type , e_value , e_tb = sys.exc_info( )
                e_string = ( 'An internal error occured while handling an '
                        + 'error.\n\n' + e_type.__name__ + ': '
                        + str( e_value ) + '\n\nTraceback:\n'
                        + ''.join( traceback.format_tb( e_tb ) ) )
                request.status = apache.HTTP_INTERNAL_SERVER_ERROR
                request.content_type = 'text/plain'
                request.write( e_string )
            finally:
                e_tb = None
        finally:
            error_data = None
        raise apache.SERVER_RETURN( apache.OK )

    def check_page( self , req , name , lang ):
        """
        This method verifies if a page is actually defined in a given language
        using both the page's identifier and the language's.
        """
        r = self.get( '/ws:data/ws:sitemap/ws:page[ @id = $pid ]' ,
                    pid = name )
        if not r:
            self.error( req , SiteXML.PageNotFound , lang = lang )
        r = self.get( '/ws:data/ws:contents[ @lang = $lid ]/ws:pages'
                        + '/ws:page[ @id = $pid ]' ,
                    pid = name , lid = lang )
        if not r:
            self.error( req , SiteXML.PageNotFound , lang = lang )



class XPathFunction( object ):
    """
    This class acts as the base class for extra XPath functions provided by
    PySSE. It contains the function registration code, as well as the call
    mechanism for calls coming from either the XPath code or from other parts
    of the PySSE script. It also includes some utilities to generate XHTML
    and PySSE tags.
    
    Classes that inherit XPathFunction must have either a get_node() method
    that returns a list of elements or a get_value() method that returns a
    scalar. They may also implement a prepare() method which will be called
    during initialisation. Finally, they must have a field named function,
    which contains the name of the new XPath function.
    """

    def __init__( self , registry ):
        """
        This method registers the function in the XPath function registry and
        prepares the function by calling the class' prepare() method (if it
        exists).
        """
        self.registry = registry
        self.site = registry.site
        registry.namespace[ self.__class__.function ] = self
        if hasattr( self , 'prepare' ):
            self.prepare( )

    def __call__( self , context , *args ):
        return self._call( *args )

    def _call( self , *args ):
        """
        This method implements actual calls to the function. It looks for a
        get_node() method, calling it if it exists; if not, it then looks for
        the get_value() method and calls it. If none of the methods exist, it
        raises the NotImplementedError exception.
        """
        if hasattr( self , 'get_node' ):
            l = self.get_node( *args )
            return [ self.ws( 'data' , tags = l ) ]
        elif hasattr( self , 'get_value' ):
            return self.get_value( *args )
        raise NotImplementedError( self.__class__.function + ' '
                + repr( args ) )

    def func( self , cls , *args ):
        """Calls another XPath function through the registry."""
        return self.registry.call( cls , *args )

    def html( self , tag , contents = None , tags = None ):
        """Generates an XHTML tag."""
        return self._tag( tag , contents , tags )

    def ws( self , tag , contents = None , tags = None ):
        """Generates a PySSE tag."""
        return self._tag( tag , contents , tags , 'ws' )

    def _tag( self , tag , contents , tags , ns = None ):
        """
        This method generates a tag. The 'tag' parameter determines the name
        of the new tag. If the 'contents' parameter is set, then it is used
        as the new node's text; otherwise, if the 'tags' parameters contains
        a list, it is used as the new tag's list of children. Finally, the
        'ns' parameter indicates the namespace to use; it must be None for
        XHTML tags or 'ws' for PySSE tags.
        """
        xhtml_ns = 'http://www.w3.org/1999/xhtml'
        ws_ns = 'http://www.ebenoit.info/data'
        nsmap = { None : xhtml_ns , 'ws' : ws_ns }
        node = etree.Element( '{%s}%s' % ( nsmap[ ns ] , tag ) ,
                nsmap = nsmap )
        if contents is not None:
            node.text = unicode( contents )
        elif tags is not None:
            node.extend( tags )
        return node



class XPathRegistry( object ):
    """This class implements the XPath functions registry."""

    def __init__( self , site ):
        """
        This method initialises the object by storing the XMLSite instance
        for the site, and creating both the function dictionnary and the XPath
        namespace.
        """
        self.site = site
        self.functions = {}
        self.namespace = etree.FunctionNamespace(
                'http://www.ebenoit.info/ext' )

    def add( self , *args ):
        """
        This method adds any number of functions (implemented by subclassing
        the XPathFunction class) to the registry.
        """
        for f_class in args:
            assert( f_class not in self.functions )
            f = f_class( self )
            self.functions[ f_class ] = f

    def call( self , cls , *args ):
        """This method calls an XPath function directly."""
        return self.functions[ cls ]._call( *args )

    def get( self , cls ):
        """This method returns the instance that implements a function."""
        return self.functions[ cls ]




class RuleSet( object ):
    """
    This class implements rulesets, which are used by page groups to determine
    whether a page is to be included or excluded from a page group.
    """

    def __init__( self , tags ):
        """
        The initialisation method of a rule set is called with the list of
        tags that define it as its parameter. It will initialise the list
        of exact and partial matches for the rule set.
        """
        self.exact = [ tag.attrib[ 'exact' ] for tag in tags
                if 'exact' in tag.attrib and tag.attrib[ 'exact' ] != '' ]
        self.partial = [ re.compile( tag.attrib[ 're' ] )
            for tag in tags
                if 're' in tag.attrib and tag.attrib[ 're' ] != '' ]


    def match( self , page ):
        """This method checks if a page matches the rules in this set."""
        if True in ( page == epn for epn in self.exact ):
            return True
        return True in ( ppn.search( page ) is not None
                    for ppn in self.partial )



class PageGroup( object ):
    """This class implements page groups."""

    def __init__( self , site , root_tag ):
        """
        The initialisation method reads a group's identifier and both sets
        of inclusion and exclusion rules from the XML tree.
        """
        # Handle group identifier
        self.id = root_tag.attrib[ 'id' ]
        if not re.match( '[a-z][a-z0-9\-]*$' , self.id , re.IGNORECASE ):
            raise ValueError( 'invalid group ID "%s"' % self.id )

        # Get inclusion rules
        r = site.get( 'ws:include/ws:match' ,
                xml_data = root_tag )
        if not r:
            raise ValueError( 'no inclusion rules in group "%s"' % self.id )
        self.include = RuleSet( r )

        # Get exclusion rules
        r = site.get( 'ws:exclude/ws:match' ,
                xml_data = root_tag )
        self.exclude = RuleSet( r )


    def in_group( self , page ):
        """This method checks if a page matches the group's rules."""
        return self.include.match( page ) and not self.exclude.match( page )




def get_languages( site ):
    """
    Returns a set of supported language identifiers. The identifiers are read
    from the main configuration file's ws:contents tags' lang attribute.
    """
    r = site.get( '/ws:data/ws:contents' )
    languages = set([ node.get( 'lang' ) for node in r ])
    if not languages:
        languages = set([ 'en' ])
    return languages



def guess_language( accept_str , languages ):
    """
    Attempts to guess the language to use from the browser's Accept-Language
    string and the set of available languages.
    """
    accept_str = re.sub( r'[^a-zA-Z0-9\.\-;,=]' , '' , accept_str )
    if len( accept_str ) == 0:
        return languages[ 0 ]

    # Build a list using the different weights as keys
    accepted = { }
    acc_list = accept_str.split( ',' )

    for lang_spec in acc_list:

        if lang_spec == '':
            continue

        lang_data = lang_spec.split( ';q=' )
        if len( lang_data ) == 1:
            lang_data.append( '1.0' )

        lang_string = lang_data[ 0 ]
        try:
            lang_weight = float( lang_data[ 1 ] ) * 1000
        except ValueError:
            continue

        if lang_weight in accepted:
            accepted[ lang_weight ].append( lang_string )
        else:
            accepted[ lang_weight ] = [ lang_string ]

    # Iterate for all languages in the weight-specific order
    weights = accepted.keys( )
    weights.sort( )
    for lang_weight in reversed( weights ) :
        for lang_string in accepted[ lang_weight ]:
            if lang_string in languages:
                return lang_string

    return languages[ 0 ]



def get_paths( req ):
    """
    Returns the various (script URI, web root, data path) from the request
    information.
    """
    # Get the script's file name
    from os.path import basename, dirname
    script_file = req.finfo[ apache.FINFO_FNAME ]

    # Get the script's URI
    script_name = basename( script_file )
    script_uri = req.parsed_uri[ apache.URI_PATH ]
    if script_uri.find( script_name ) == -1:
        script_uri += script_name
    elif req.path_info:
        script_uri = script_uri[ 0 : -len( req.path_info ) ]

    # Get the site's base URI
    base_uri = dirname( script_uri )
    if base_uri[ len( base_uri ) - 1 ] != '/' :
        base_uri = base_uri + '/'

    # Get the data path
    data_path = dirname( script_file ) + '/data/'

    return ( script_uri , base_uri , data_path )


def set_content_type( req ):
    """
    This function generates the proper Content-Type header in the response,
    depending on what the client accepts.
    """
    if 'Accept' in req.headers_in:
        accept = re.split( '[;,]' , req.headers_in[ 'Accept' ].lower() )
        if 'application/xhtml+xml' in accept:
            req.content_type = 'application/xhtml+xml'
        else:
            req.content_type = 'text/html'
    else:
        req.content_type = 'text/html'



##############################################################################
# XPATH FUNCTIONS                                                            #
##############################################################################

class GetI18NString( XPathFunction ):
    """Implementation of the wsf:get-i18n-string XPath function."""

    function = 'get-i18n-string'

    def prepare( self ):
        """
        This method prepares the two caches used by the function - string
        files and actual strings.
        """
        self.translations = {}
        self.translation_files = {}

    def get_value( self , string_id , lang_id ):
        """
        Looks up a string from the translation tables. If the string had been
        requested in the past, return the cached version. If not, first look
        for this string in the configuration file, and if it is not there,
        attempt to retrieve it from the appropriate translation file.
        """
        # Is it cached?
        if ( string_id , lang_id ) in self.translations:
            return self.translations[ ( string_id , lang_id ) ]

        # Look in the main configuration file
        xpath_root = "/ws:data/ws:contents[ @lang = $lid ]/ws:strings"
        xpath_expr = xpath_root + "/ws:string[ @id = $sid ]"
        r = self.site.get( xpath_expr , lid = lang_id , sid = string_id )
        if len( r ) >= 1:
            self.translations[ ( string_id , lang_id ) ] = r[ 0 ].text
            return self.translations[ ( string_id , lang_id ) ]

        # Is there a separate data file?
        if lang_id not in self.translation_files:
            xpath_expr = xpath_root + "[ @datafile ]"
            r = self.site.get( xpath_expr , lid = lang_id )
            if len( r ) < 1:
                return unicode( '(string not found)' )
            self.translation_files[ lang_id ] = etree.parse(
                    self.site.path + r[0].attrib[ 'datafile' ]
                )

        # Look in the separate file
        f = self.translation_files[ lang_id ]
        r = self.site.get( '/ws:strings/ws:string[ @id = $sid ]' ,
                           xml_data = f , sid = string_id )
        if len( r ) < 1:
            return unicode( '(string not found)' )
        self.translations[ ( string_id , lang_id ) ] = r[ 0 ].text
        return self.translations[ ( string_id , lang_id ) ]



class GetPageTitle( XPathFunction ):
    """Implementation of the wsf:get-page-title XPath function."""

    function = 'get-page-title'

    def prepare( self ):
        """Initialises the cache"""
        self.titles = {}

    def lookup_short( self , page_id , lang_id ):
        """
        This method looks up the short title for a page based on its
        identifier and language identifier. If no title is found, KeyError
        is raised.
        """
        # Look in the language-specific section
        xpath_expr = ( "/ws:data/ws:contents[ @lang = $lid ]/ws:pages"
                     + "/ws:page[ @id = $pid and @title ]" )
        r = self.site.get( xpath_expr , lid = lang_id , pid = page_id )
        if len( r ) >= 1:
            return r[ 0 ].attrib[ 'title' ]
        # Look in the sitemap
        xpath_expr = "/ws:data/ws:sitemap/ws:page[ @id = $pid and @title ]"
        r = self.site.get( xpath_expr , pid = page_id )
        if len( r ) >= 1:
            return r[ 0 ].attrib[ 'title' ]
        raise KeyError

    def get_long_title( self , page_id , lang_id ):
        """
        This method returns the details about a page's long title, based on
        its identifier and language. It also verifies that the long title
        mode is correct (i.e. no replace mode for groups or languages).
        If no long title is found, KeyError is raised.
        """
        # Create the lists of paths that should be examined
        contents_pfx = '/ws:contents[ @lang = $lid ]/'
        groups = self.site.get_groups( page_id )
        xpaths = [ ( True ,
                contents_pfx + 'ws:pages/ws:page[ @id = $pid ]' ) ]
        for g_id in groups:
            xpaths.append( ( False , contents_pfx + "ws:group[ @id = '"
                    + g_id + "' ]" ) )
        xpaths.append( ( True , "/ws:sitemap/ws:page[ @id = $pid ]" ) )
        for g_id in groups:
            xpaths.append( ( False , "/ws:group[ @id = '" + g_id + "' ]" ) )
        xpaths.append( ( False , '' ) )

        # Go through the list of paths
        for replace_ok , xp_base in xpaths:
            xpath = '/ws:data' + xp_base + '/ws:long-title'
            r = self.site.get( xpath , lid = lang_id , pid = page_id )
            if len( r ) != 1:
                continue
            tag = r[ 0 ]
            # Get the mode
            if 'mode' in tag.attrib:
                mode = tag.attrib[ 'mode' ]
            elif replace_ok:
                mode = 'replace'
            else:
                mode = 'prefix'
            # Check it and return
            if mode not in ( 'replace' , 'prefix' , 'suffix' ):
                raise ValueError( 'invalid long title mode "%s"' % mode )
            elif mode == 'replace' and not replace_ok:
                raise ValueError(
                    'long title mode "%s" not permitted for page "%s"'
                        % ( mode , page_id ) )
            return ( mode == 'replace' , mode == 'prefix' , tag.text )

        # Failed
        raise KeyError

    def make_long_title( self , page_id , lang_id ):
        """
        This method tries to generate a page's long title. If the page has no
        long title settings, the short title is returned.
        """
        # Get long title settings
        try:
            t_data = self.get_long_title( page_id , lang_id )
        except KeyError:
            return self.lookup_short( page_id , lang_id )
        # Replace short title?
        if t_data[ 0 ]:
            return t_data[ 2 ]
        # Prefix/suffix modes
        t = self.get_value( page_id , lang_id )
        if t_data[ 1 ]:
            t = t_data[ 2 ] + t
        else:
            t = t + t_data[ 2 ]
        return t


    def get_value( self , page_id , lang_id , long_title = False ):
        """
        This method looks up a page's title in the correct language, using
        the correct type (normal or long) of title.
        """
        t_id = ( page_id , lang_id , long_title )
        if t_id not in self.titles:
            try:
                if long_title:
                    t = self.make_long_title( page_id , lang_id )
                else:
                    t = self.lookup_short( page_id , lang_id )
            except KeyError:
                t = unicode( '(title not found)' )
            self.titles[ t_id ] = t
        return self.titles[ t_id ]



class GetPageAuthor( XPathFunction ):
    """Implementation of the wsf:get-page-author XPath function."""

    function = 'get-page-author'

    def prepare( self ):
        """Initialises the cache."""
        self.authors = {}

    def get_value( self , page_id , lang_id ):
        """
        Looks up the name of a page's author from the configuration file.
        First, the cache is examined and its contents returned if there is
        a match. If not, the language- and page-specific author name is looked
        up; if it fails, the page-specific author is then looked up. Finally,
        if that last attempt failed as well, the site-wise author name will
        be returned.
        """
        # Is it cached?
        if ( page_id , lang_id ) in self.authors:
            return self.authors[ ( page_id , lang_id ) ]

        # Create the lists of paths that should be examined
        contents_pfx = '/ws:contents[ @lang = $lid ]/'
        groups = self.site.get_groups( page_id )
        xpaths = [ contents_pfx + 'ws:pages/ws:page[ @id = $pid ]' ]
        for g_id in groups:
            xpaths.append( contents_pfx + "ws:group[ @id = '" + g_id
                    + "' ]" )
        xpaths.append( "/ws:sitemap/ws:page[ @id = $pid ]" )
        for g_id in groups:
            xpaths.append( "/ws:group[ @id = '" + g_id + "' ]" )
        xpaths.append( '' )

        # Go through the list of path and return the first name we find
        for xp_base in xpaths:
            xpath = '/ws:data' + xp_base + '/ws:author'
            r = self.site.get( xpath , lid = lang_id , pid = page_id )
            if len( r ) >= 1:
                self.authors[ ( page_id , lang_id ) ] = r[ 0 ].text
                break
        if ( page_id , lang_id ) not in self.authors:
            self.authors[ ( page_id , lang_id ) ] = ''
        return self.authors[ ( page_id , lang_id ) ]



class GetPageData( XPathFunction ):
    """Implementation of the wsf:get-page-data XPath function"""

    function = 'get-page-data'

    def prepare( self ):
        """Initialises the cache"""
        self.page_data = {}

    def get_node( self , page_id , lang_id ):
        """
        This method attempts to find and return the contents of a page based
        on its identifier and on the requested language. The data is returned
        from either the configuration file itself or the specified data file.
        """
        # Is it cached?
        if ( page_id , lang_id ) in self.page_data:
            return self.page_data[ ( page_id , lang_id ) ]

        # Find the page's record
        xpath_expr = ( '/ws:data/ws:contents[ @lang = $lang ]'
                       + '/ws:pages/ws:page[ @id = $pid ]' )
        page_lst = self.site.get( xpath_expr , lang = lang_id , pid = page_id )
        if len( page_lst ) != 1:
            raise ValueError( 'Page "%s" in language "%s" not found'
                                    % ( page_id , lang_id ) )
        page = page_lst[ 0 ]

        # Does it have a separate data file?
        if 'datafile' not in page.attrib:
            # Nope, return the tag's contents
            self.page_data[ ( page_id , lang_id ) ] = list( page )
            return self.page_data[ ( page_id , lang_id ) ]

        # Determine the path to the page's data file
        file_path = self.site.path

        contents_pfx = "/ws:contents[ @lang = $lid ]"
        groups = self.site.get_groups( page_id )
        xpaths = [ contents_pfx + "/ws:group[ @id = '" + g_id + "' ]"
                for g_id in groups ]
        xpaths.append( contents_pfx )
        xpaths.extend( [ "/ws:group[ @id = '" + g_id + "' ]"
                for g_id in groups ] )
        xpaths.append( '' )
        for xp_base in xpaths:
            xpath = '/ws:data' + xp_base + '/ws:page-data'
            r = self.site.get( xpath , lid = lang_id )
            if len( r ) == 1:
                file_path += r[ 0 ].text + '/'
                break
        file_path += page.attrib[ 'datafile' ] + '.xml'

        # Get the data from the file
        page_file = etree.parse( file_path )
        self.page_data[ ( page_id , lang_id ) ] = list( page_file.getroot() )
        return self.page_data[ ( page_id , lang_id ) ]



class GetSideBar( XPathFunction ):
    """Implementation of the wsf:get-side-bar XPath function."""

    function = 'get-side-bar'

    def prepare( self ):
        """Initialise the cache"""
        self.sidebars = {}

    def get_node( self , page ):
        """Returns the side bar configuration that applies to a page."""
        # Did we cache it?
        if page in self.sidebars:
            return self.sidebars[ page ]

        # Try to find the definition
        xpath_expr = "/ws:data/ws:side-bar[ @page = $pid ]"
        r = self.site.get( xpath_expr , pid = page )
        if len( r ) == 0:
            # No page-specific configuration, try groups
            gx = "' or @group='".join( self.site.get_groups( page ) )
            r = self.site.get( "/ws:data/ws:side-bar[ @group = '"
                    + gx + "' ]" )
            if len( r ) == 0:
                r = self.site.get( "/ws:data/ws:side-bar[ not( @page ) ]" )

        # Cache result and return
        if len( r ) == 0:
            self.sidebars[ page ] = [ ]
        else:
            self.sidebars[ page ] = [ r[ 0 ] ]
        return self.sidebars[ page ]



class GenerateBreadcrumb( XPathFunction ):
    """Implementation of the wsf:generate-breadcrumb XPath function."""

    function = 'generate-breadcrumb'

    def get_node( self , page_id , lang_id , omit_last , top_level_string ):
        """
        This function generates the nodes for a breadcrumb, given a page
        identifier, a language, and two additional parameters. The first of
        these parameters, omit_last, indicates that the last entry in the
        breadcrumb (that is, the page itself) should not be displayed. The
        top_level_string parameter indicates that the breadcrumb-top string
        should be displayed instead of the breadcrumb when the current page
        is the site's home page.
        """
        # Find out the home page's identifier
        r = self.site.get( '/ws:data/ws:sitemap' )
        if 'home' in r[ 0 ].attrib:
            home_id = r[ 0 ].attrib[ 'home' ]
        else:
            home_id = 'main'

        if page_id == home_id:
            # This is the root, do we display the string?
            if top_level_string:
                span_text = self.func( GetI18NString , 'breadcrumb-top' ,
                        lang_id )
            else:
                span_text = self.func( GetPageTitle , page_id , lang_id )
            return [ self.html( 'span' , contents = span_text ) ]

        # Get the breadcrumb separator
        r = self.site.get( '/ws:data/ws:breadcrumb[ @separator ]' )
        if len( r ) >= 1:
            bc_sep = r[ 0 ].attrib[ 'separator' ]
        else:
            bc_sep = unicode( ' ' )

        # Not at the root
        bc = [ ]
        current = page_id
        previous = ''
        while True:

            # Create the crumb
            if not( current == page_id and omit_last ):
                n_text = self.func( GetPageTitle , current , lang_id )
                if previous != page_id or not( omit_last ):
                    if current == page_id:
                        span_text = n_text
                    else:
                        span_text = bc_sep
                    bc.insert( 0 , self.html( 'span' , span_text ) )

                if current != page_id:
                    # This is a link
                    link = self.ws( 'link-to' , n_text )
                    link.attrib[ 'page' ] = current
                    bc.insert( 0 , link )

            # End the loop if that was the home page
            if current == home_id:
                break

            # Get the previous page in the hierarchy
            previous = current
            xpath_expr = ( "/ws:data/ws:sitemap/ws:page"
                         + "[ @id = $pid and @parent ]" )
            r = self.site.get( xpath_expr , pid = current )
            if len( r ) >= 1:
                current = r[ 0 ].attrib[ 'parent' ]
            else:
                current = home_id

        return bc



class GetMetaData( XPathFunction ):
    """Implements the wsf:get-meta-data XPath function."""

    function = 'get-meta-data'

    def prepare( self ):
        """Initialise the cache."""
        self.cache = {}

    def get_node( self , lang , page , tag , id_field = None ):
        """
        This function returns a list of values for a metadata or inclusion
        tag, on a given page in a specific language. If the tag to look
        up uses an identifier, the name of the attribute which contains it
        is also passed as a parameter.
        """
        # Check if the specified tag is supported
        if tag not in ( 'meta' , 'script' , 'css' , 'image-preload' ,
                'keyword' ):
            raise ValueError( 'unknown tag "%s"' % tag )

        # Is the data cached?
        meta_id = ( tag , lang , page )
        if meta_id in self.cache:
            return self.cache[ meta_id ]

        # List of XPath expressions for the tag
        x_exprs = [ '' ]
        for g_id in self.site.get_groups( page ):
            x_exprs.append( '/ws:group[ @id = \'' + g_id + '\' ]' )
        x_exprs.append( '/ws:contents[ @lang = $lang ]' )
        for g_id in self.site.get_groups( page ):
            x_exprs.append( '/ws:contents[ @lang = $lang ]'
                    + '/ws:group[ @id = \'' + g_id + '\' ]' )
        x_exprs.extend( [
                '/ws:sitemap/ws:page[ @id = $page ]' ,
                '/ws:contents[ @lang = $lang ]/ws:pages/ws:page[ @id = $page ]'
            ] )

        # Depending on whether there is an id field or not, generate a list or
        # a dictionnary
        if id_field is None:
            data = [ ]
            values = [ ]
        else:
            data = { }
            values = None

        # Iterate through the list of expressions
        for xe_base in x_exprs:
            xe = '/ws:data' + xe_base + ( '/ws:%s' % tag )
            r = self.site.get( xe , lang = lang , page = page )

            # Iterate through the results
            for t in r:

                if id_field is None:
                    # Handle lists
                    value = t.text
                    if value not in values:
                        data.append( deepcopy( t ) )
                        values.append( value )
                    continue

                if id_field not in t.attrib:
                    # Ignore tags that are missing their identifier
                    continue

                # Handle tags with identifiers
                data[ t.attrib[ id_field ] ] = deepcopy( t )

        # If there was an identifier, generate the list
        if id_field is not None:
            data = data.values( )

        # Cache the data and return it
        self.cache[ meta_id ] = data
        return data



class GetDataPath( XPathFunction ):
    """Implements the wsf:get-data-uri XPath function."""

    function = 'get-data-uri'

    def prepare( self ):
        """Initialises the cache"""
        self.cache = {}

    def get_value( self , d_type ):
        """
        Looks up the URI to a given data type from the configuration file.
        """
        if d_type not in ( 'js' , 'css' , 'img' ):
            raise ValueError( 'Invalid data type "%s"' % d_type )
        if d_type not in self.cache:
            r = self.site.get(
                    '/ws:data/ws:data-paths/ws:path-to[ @type = $id ]' ,
                    id = d_type )
            if len( r ) >= 1:
                self.cache[ d_type ] = r[ 0 ].text
            else:
                self.cache[ d_type ] = d_type
        return self.cache[ d_type ]



class GetSiteTitle( XPathFunction ):
    """Implements the wsf:get-site-title XPath function."""

    function = 'get-site-title'

    def prepare( self ):
        """Initialises the cache."""
        self.cache = {}

    def get_value( self , lang ):
        """
        Looks up the site's title in the specified language, and if there
        is no language-specific title, use the site-wise title.
        """
        if lang in self.cache:
            return self.cache[ lang ]

        r = self.site.get( '/ws:data/ws:contents[ @lang = $lang ]/ws:title' ,
                lang = lang )
        if len( r ) == 0:
            r = self.site.get( '/ws:data/ws:title' )
            if len( r ) == 0:
                self.cache[ lang ] = self.func( GetI18NString ,
                        'untitled-site' , lang )
        if lang not in self.cache:
            self.cache[ lang ] = r[ 0 ].text

        return self.cache[ lang ]



class MakeLinkClass( XPathFunction ):
    """Implements the wsf:make-link-class XPath function."""

    function = 'make-link-class'

    def prepare( self ):
        """Initialises the various configuration fields and the cache."""
        self.link_debug = None
        self.page_not_found = None
        self.anchor_not_found = None
        self.cache = { }

    def get_config( self ):
        """
        Reads the configuration for the link debugging system and fill
        the missing entries using default values.
        """
        r = self.site.get( '/ws:data/ws:link-debug' )
        if ( len( r ) != 1 or 'enable' not in r[ 0 ].attrib or
                r[ 0 ].attrib[ 'enable' ] != '1' ):
            self.link_debug = False
            return

        self.link_debug = True
        tag = r[ 0 ]
        if 'no-page' in tag.attrib:
            self.page_not_found = tag.attrib[ 'no-page' ]
        else:
            self.page_not_found = 'pysse-ld-page'
        if 'no-anchor' in tag.attrib:
            self.anchor_not_found = tag.attrib[ 'no-anchor' ]
        else:
            self.anchor_not_found = 'pysse-ld-anchor'

    def check_anchor( self , lang , page , anchor ):
        """
        This method checks if an anchor exists in the body of a page (in a
        specific language). It returns either an empty string if the anchor
        was found or the configuration value for missing anchors if it
        wasn't.
        """
        try:
            contents = self.ws( 'data' , tags = self.func(
                                GetPageData , page , lang ) )
        except:
            return self.page_not_found
        
        r = self.site.get( '//*[ @id = $anchor ]' , xml_data = contents ,
                anchor = anchor )
        if len( r ) == 1:
            return ''
        return self.anchor_not_found

    def get_value( self , lang , page , anchor , existing ):
        """
        This method checks if a link's target exists, and returns
        a combination of the pre-defined class(es) and the link debugging
        class, depending on the link's status.
        """
        # Check if link debugging is active
        if self.link_debug is None:
            self.get_config( )

        # If this is a relative link or if debugging is disabled,
        # return the existing value
        if page == '' or self.link_debug is False:
            return existing

        # If the value has been cached, return it
        p_id = ( lang , page , anchor )
        if p_id in self.cache:
            return ( existing + ' ' + self.cache[ p_id ] ).strip( )

        # Does the page exist in the sitemap?
        r = self.site.get( '/ws:data/ws:sitemap/ws:page[ @id = $page ]' ,
                page = page )
        if len( r ) != 1:
            self.cache[ p_id ] = self.page_not_found
        else:
            # Does it exist in the specified language?
            r = self.site.get( '/ws:data/ws:contents[ @lang = $lang]'
                                + '/ws:pages/ws:page[ @id = $page ]' ,
                               lang = lang , page = page )
            if len( r ) != 1:
                self.cache[ p_id ] = self.page_not_found
            elif anchor == '':
                self.cache[ p_id ] = ''
            else:
                self.cache[ p_id ] = self.check_anchor( lang , page ,
                        anchor )

        return ( existing + ' ' + self.cache[ p_id ] ).strip( )



class GenerateMenu( XPathFunction ):
    """Implements the wsf:generate-menu XPath function."""

    function = 'generate-menu'

    DEFAULT_DEPTH = 2

    def prepare( self ):
        """Initialise the cache."""
        self.cache = { }
        self.max_depth = None

    def get_max_depth( self ):
        """
        This method attempts to read the menu's maximal depth from the
        configuration file; if the value can't be found, it is set to the
        default value.
        """
        r = self.site.get( '/ws:data/ws:menu[ @depth ]' )
        if len( r ) != 1:
            self.max_depth = GenerateMenu.DEFAULT_DEPTH
        else:
            try:
                self.max_depth = int( r[ 0 ].attrib[ 'depth' ] )
                if self.max_depth < 1:
                    self.max_depth = GenerateMenu.DEFAULT_DEPTH
            except ValueError:
                self.max_depth = GenerateMenu.DEFAULT_DEPTH

    def make_auto_menu( self , lang , page , cls ):
        """
        This method generates the automatic menu for a given page in a
        specific language, using the specified class to look up the XHTML
        elements to be used as anchors. It returns a list of ws:menu-item
        tags.
        """
        items = [ ]
        try:
            contents = self.ws( 'data' , tags = self.func(
                                GetPageData , page , lang ) )
        except:
            return items
        
        r = self.site.get( '//*[ @class = $cls and @id ]' ,
                xml_data = contents , cls = cls )
        for tag in r:
            lnk = self.ws( 'link-to' , contents = tag.text )
            lnk.attrib[ 'page' ] = page
            lnk.attrib[ 'anchor' ] = tag.attrib[ 'id' ]
            lnk.attrib[ 'title' ] = ( self.func( GetPageTitle , page , lang )
                    + ' - ' + tag.text )
            items.append( self.ws( 'menu-item' , tags = [ self.ws( 'content' ,
                    tags = [ lnk ]
                ) ] ) )

        return items

    def make_menu( self , lang , current , parent = None , parsed = None ,
            depth = 1 ):
        """
        This method is responsible for generating the menu. Its two mandatory
        parameters, lang and current, indicate the language and the current
        active page. In addition, the parent parameter specifies the
        identifier of the parent page (or is set to None to build the menu's
        top level); the parsed parameter, if not None, is a list of node
        identifiers that were already handled. Finally, the depth parameter
        indicates the current depth being generated. The method returns a
        tuple of two values - a boolean indicating whether the list of items
        contain the current page, and the list of items.
        """
        # Create the list of already parsed nodes if needed
        if parsed is None:
            parsed = []

        # Get the list of nodes that apply
        if parent is None:
            nodes = self.site.get(
                    '/ws:data/ws:sitemap/ws:page[ not( @parent ) ]' )
        else:
            nodes = self.site.get(
                    '/ws:data/ws:sitemap/ws:page[ @parent = $p ]' ,
                    p = parent )

        # Iterate through the nodes to generate the menu
        items = [ ]
        active_here = False
        for node in nodes:
            # Ignore nodes without ID and hidden nodes
            if 'id' not in node.attrib or 'hidden' in node.attrib:
                continue

            # If a node seems to have been parsed once in the past, skip it
            id = node.attrib[ 'id' ]
            if id in parsed:
                continue
            parsed.append( id )
            is_current = ( id == current )
            if is_current:
                active_here = True

            # If we are already below maximal depth, we're only looking
            # for the current page
            if depth > self.max_depth:
                if id != current:
                    active_here, sub = self.make_menu( lang , current , id ,
                            parsed , depth + 1 )
                if active_here:
                    break
                continue

            # Determine if a sub-menu is to be generated
            sub = None
            if is_current and depth == self.max_depth:
                sub = [ ]
                contains_active = False
            elif depth == self.max_depth:
                if not active_here:
                    contains_active = self.make_menu( lang , current , id ,
                            parsed , depth + 1 )[ 0 ]
                sub = [ ]

            # Generate sub-menu, it that applies
            if sub is None:
                contains_active , sub = self.make_menu( lang , current , id ,
                        parsed , depth + 1 )
                if not sub and 'auto-menu' in node.attrib:
                    sub = self.make_auto_menu( lang , id ,
                            node.attrib[ 'auto-menu' ] )

            # Make sure active_here is up-to-date
            active_here = active_here or contains_active

            # Generate contents
            text = self.func( GetPageTitle , id , lang )
            if is_current:
                contents = self.html( 'strong' , contents = text )
            else:
                contents = self.ws( 'link-to' , contents = text )
                contents.attrib[ 'page' ] = id
            c_tag = self.ws( 'content' , tags = [ contents ] )

            # Generate the menu tag
            m_tag_sub = [ c_tag ]
            if sub:
                m_tag_sub.append( self.ws( 'sub-menu' , tags = sub ) )
            m_tag = self.ws( 'menu-item' , tags = m_tag_sub )
            if sub and ( contains_active or is_current ):
                m_tag.attrib[ 'open' ] = '1'
            items.append( m_tag )

        return active_here , items


    def get_node( self , lang , current ):
        """Main menu generation method."""
        # Use cached data if possible
        if ( lang , current ) not in self.cache:
            # Read maximal depth from the configuration
            if self.max_depth is None:
                self.get_max_depth( )
            # Generate the menu
            self.cache[ ( lang , current ) ] = self.make_menu(
                    lang , current )[ 1 ]
        return self.cache[ ( lang , current ) ]



class GeneratePageMenu( XPathFunction ):
    """This class implements the wsf:get-page-menu XPath function."""

    function = 'generate-page-menu'

    def prepare( self ):
        """Initialise the configuration."""
        self.enabled = False
        self.counter = 0
        self.min_level = 1
        self.max_level = 6


    def read_configuration( self ):
        """This method reads the page menu generator's configuration."""
        r = self.site.get( '/ws:data/ws:on-this-page' )
        if len( r ) != 1:
            return
        cfg = r[ 0 ]

        if 'min-level' in cfg.attrib:
            try:
                self.min_level = int( cfg.attrib[ 'min-level' ] )
            except ValueError:
                pass

        if 'max-level' in cfg.attrib:
            try:
                self.max_level = int( cfg.attrib[ 'max-level' ] )
            except ValueError:
                pass


    def get_node( self , lang , page ):
        """
        This method generates the page menu's structure for a specified page.
        It returns a structure similar to that generated by the site's menu.
        """
        # Get configuration
        if not self.enabled:
            self.enabled = True
            self.read_configuration( )

        # Get the page's contents
        contents = self.ws( 'data' , tags = self.func( GetPageData ,
                page , lang ) )

        # Generate the XPath expression
        xpath_expr = '|'.join( ( ( '//xhtml:h%d' % i)
                    for i in range( self.min_level, self.max_level + 1 ) ) )
        r = self.site.get( xpath_expr , xml_data = contents )

        stack = []
        items = [ ]

        def _stack_pop( ):
            s_top = stack.pop( )
            if s_top[ 2 ]:
                s_top[ 1 ].append( self.ws( 'sub-menu' ,
                        tags = s_top[ 2 ] ) )
            if stack:
                stack[ -1 ][ 2 ].append( s_top[ 1 ] )
            else:
                items.append( s_top[ 1 ] )

        # Generate menu
        for tag in r:
            if 'id' in tag.attrib:
                id = tag.attrib[ 'id' ]
            else:
                self.counter = self.counter + 1
                id = 'auto-h-identifier-%d' % self.counter
            link_tag = self.html( 'a' , contents = tag.text )
            link_tag.attrib[ 'href' ] = '#' + id
            link_tag.attrib[ 'title' ] = tag.text
            mi_tag = self.ws( 'menu-item' , tags = [ self.ws( 'content' ,
                    tags = [ link_tag ] ) ] )
            mi_tag.attrib[ 'open' ] = '1'

            level = int( tag.tag[ -1 ] )
            while stack:
                if stack[ -1 ][ 0 ] >= level:
                    _stack_pop( )
                else:
                    break
            stack.append( ( level , mi_tag , [ ] ) )

        while stack:
            _stack_pop( )

        return items



class GenerateHIdentifier( XPathFunction ):
    """This class implements the wsf:generate-h-identifier XPath function."""

    function = 'generate-h-identifier'

    def prepare( self ):
        """
        Initialises the counter and marks the activation status as unknown.
        """
        self.counter = 0
        self.enabled = None

    def get_value( self , tag , id ):
        """
        This method generates the identifier for a <hX> HTML tag, based on
        the page menu's configuration. Existing identifiers are returned
        directly.
        """

        # If there was already an identifier, return it
        if id != '':
            return id

        # Get configuration from the GeneratePageMenu function
        if self.enabled is None:
            gpm = self.registry.get( GeneratePageMenu )
            self.enabled = gpm.enabled
            self.min_level = gpm.min_level
            self.max_level = gpm.max_level

        # Check if the generator is to be enabled
        if not self.enabled or len( tag ) != 2 or tag[ 0 ] != 'h':
            return ''
        try:
            level = int( tag[ 1 ] )
        except ValueError:
            return ''
        if level not in range( self.min_level , self.max_level + 1 ):
            return ''

        # Generate the next identifier
        self.counter = self.counter + 1
        return 'auto-h-identifier-%d' % self.counter




##############################################################################
# REQUEST HANDLER                                                            #
##############################################################################

def handler( req ):

    # Find the script's base URI and physical path
    script_uri , base_uri , data_path = get_paths( req )

    # Define the site, then register XPath functions
    site = SiteXML( data_path )
    registry = XPathRegistry( site )
    registry.add( GetI18NString , GetPageTitle , GetPageAuthor ,
            GetPageData , GetSideBar , GenerateBreadcrumb , GetMetaData ,
            GetDataPath , GetSiteTitle , MakeLinkClass , GenerateMenu ,
            GeneratePageMenu , GenerateHIdentifier )

    # Attempt to read the site's XML file, and display error pages accordingly
    try:
        site.xml
    except IOError:
        site.error( req , SiteXML.NoConfiguration )
    except etree.XMLSyntaxError:
        site.error( req , SiteXML.ConfigurationError , sys.exc_info() )

    # The only method we accept is "GET"
    if req.method != 'GET':
        site.error( req , SiteXML.BadMethod )

    # Find out which languages are supported
    languages = list( get_languages( site ) )

    # Extract requested language and page from the path
    r_lang = ""
    r_page = "main"
    if req.path_info:
        splitPath = req.path_info[ 1 : ].split( '/' )
        if len( splitPath ) == 1:
            r_page = splitPath.pop(0)
        elif len( splitPath ) >= 2:
            r_lang = splitPath.pop(0)
            r_page = splitPath.pop(0)

    # Check if the selected language is valid
    if r_lang is not None and r_lang not in languages:
        r_lang = None

    # If there is no specified language, try to guess it
    # according to Accept-Language
    if r_lang is None:
        if req.headers_in.has_key( 'Accept-Language' ):
            accept_string = req.headers_in[ 'Accept-Language' ]
            r_lang = guess_language( accept_string , languages )
        else:
            r_lang = languages[ 0 ]

    # Check if the page name is valid
    if not re.match( '[a-z][a-z0-9\-]*$' , r_page , re.IGNORECASE ):
        site.error( req , SiteXML.PageNotFound , lang = r_lang )

    # Check if the page actually exists
    site.check_page( req , r_page , r_lang )

    # Load page groups
    site.init_groups( )

    # Load the XSLT stylesheet
    try:
        transform = etree.XSLT( etree.parse( data_path + 'site.xsl' ) )
    except ( IOError , etree.XMLSyntaxError ):
        site.error( req , SiteXML.ConfigurationError , sys.exc_info() ,
                r_lang )

    # Generate the page
    try:
        page = str( transform( site.xml ,
                reqPage = "'" + r_page + "'" ,
                reqLang = "'" + r_lang + "'" ,
                webRoot = "'" + base_uri + "'" ,
                scriptPath = "'" + script_uri + "'" ) )
    except Exception:
        site.error( req , SiteXML.PageError , sys.exc_info() , r_lang )

    # Return the generated contents
    set_content_type( req )
    req.write( page )

    return apache.OK
